from django.test import TestCase
from .mixin import FollowMixin
from .models import Action
from django.contrib.auth.models import User
import random
import string
import time


def get_random(N=10):
    return ''.join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(N))


class FollowUser(FollowMixin, User):
    class Meta:
        proxy = True


class FollowTestCase(TestCase):
    def setUp(self):
        self.user1 = FollowUser.objects.create_user('1')
        self.user2 = FollowUser.objects.create_user('2')
        self.user3 = FollowUser.objects.create_user('3')

    def test_follow(self):
        self.user1.follow(self.user2)
        self.assertTrue(self.user1.is_following(self.user2))
        self.user1.follow(self.user2)  # Should do nothing
        self.assertTrue(self.user1.is_following(self.user2))
        self.assertEquals(self.user1.get_followings_count(), 1)
        self.assertEquals(self.user2.get_followings_count(), 0)
        self.assertEquals(self.user1.get_followers_count(), 0)
        self.assertEquals(self.user2.get_followers_count(), 1)

    def test_multiple_follow(self):
        self.user1.follow(self.user2)
        self.user1.follow(self.user3)
        self.assertEquals(self.user1.get_followings_count(), 2)

    def test_action(self):
        Action.create_action(self.user1, 'joins')
        Action.create_action(self.user1, 'follows', self.user2)

    def test_timeline(self):
        self.user1.follow(self.user2)
        Action.create_action(self.user2, 'joins')
        Action.create_action(self.user2, 'insults', self.user1)
        Action.create_action(self.user3, 'ignores', self.user1)
        actions = self.user1.get_timeline()
        self.assertAlmostEquals(actions.count(), 2)

    def test_timeline_performance(self):
        friends = 500
        actions_per_friend = 100
        paginate_actions = 15
        for i in range(friends):
            user = FollowUser.objects.create_user(get_random())
            self.user1.follow(user)
            for j in range(actions_per_friend):
                Action.create_action(user, 'insults', self.user3)
        start = time.time()
        actions = self.user1.get_timeline(count=paginate_actions)
        for action in actions:
            x = action.actor
        run_time = time.time() - start
        print(run_time)
        self.assertLess(run_time, 1.0)



from rest_framework import viewsets
from rest_framework.pagination import CursorPagination
from .serializers import ActionSerializer
from .models import Action


class ActionCursorPagination(CursorPagination):
    ordering = 'id'


class ActionViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Action.objects.all().select_related().prefetch_related(
        'target', 'actor')
    serializer_class = ActionSerializer
    pagination_class = ActionCursorPagination
    filter_fields = (
        'actor_content_type', 'actor_content_type__model',
        'target_content_type', 'target_content_type__model',
    )



from .models import Follow, Action
from django.contrib.contenttypes.models import ContentType


def get_content_type(obj):
    return ContentType.objects.get_for_model(obj)


class FollowMixin(object):
    def _get_follow(self, followee):
        try:
            follow = Follow.objects.get(
                follower_content_type=get_content_type(self),
                follower_object_id=self.id,
                followee_content_type=get_content_type(followee),
                followee_object_id=followee.id,
            )
            return follow
        except Follow.DoesNotExist:
            return None

    def follow(self, followee):
        follow = self._get_follow(followee)
        if follow is None:
            follow = Follow(follower=self, followee=followee)
            follow.save()
        return follow

    def unfollow(self, followee):
        follow = self._get_follow(followee)
        if follow is not None:
            follow.delete()

    def is_following(self, followee):
        follow_count = Follow.objects.filter(
            follower_content_type=get_content_type(self),
            follower_object_id=self.pk,
            followee_content_type=get_content_type(followee),
            followee_object_id=followee.pk,
        ).count()
        if follow_count > 0:
            return True
        return False

    def get_followers(self):
        return Follow.objects.filter(
            followee_content_type=get_content_type(self),
            followee_object_id=self.pk,
        )

    def get_followings(self, target_content_type=None):
        return Follow.objects.filter(
            follower_content_type=get_content_type(self),
            follower_object_id=self.pk,
        )

    def get_followers_count(self):
        return self.get_followers().count()

    def get_followings_count(self):
        return self.get_followings().count()

    def get_followings_ids(self):
        return self.get_followings().values_list(
            'followee_object_id', flat=True)

    def get_timeline(self, myself=False, verb=None, count=None, time=None):
        if myself is True:
            ids = [self.pk]
        else:
            ids = self.get_followings_ids()
        ct = get_content_type(self)
        actions = Action.get_timeline(verb=verb, time=time)
        actions = actions.filter(
            actor_object_id__in=ids,
            actor_content_type=ct,
        )
        if count is not None:
            actions = actions[:count]
        return actions

    def create_action(self, verb, target=None, timestamp=None):
        return Action.create_action(
            self, verb, target=target, timestamp=timestamp)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import datetime


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(primary_key=True, auto_created=True, serialize=False, verbose_name='ID')),
                ('actor_object_id', models.PositiveIntegerField()),
                ('verb', models.CharField(max_length=255)),
                ('target_object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('timestamp', models.DateTimeField(default=datetime.datetime.today)),
                ('actor_content_type', models.ForeignKey(to='contenttypes.ContentType')),
                ('target_content_type', models.ForeignKey(to='contenttypes.ContentType', blank=True, null=True, related_name='target_action_set')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Follow',
            fields=[
                ('id', models.AutoField(primary_key=True, auto_created=True, serialize=False, verbose_name='ID')),
                ('follower_object_id', models.PositiveIntegerField()),
                ('followee_object_id', models.PositiveIntegerField()),
                ('is_mutual', models.BooleanField(default=False)),
                ('followee_content_type', models.ForeignKey(to='contenttypes.ContentType', related_name='followee_follow_set')),
                ('follower_content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='follow',
            unique_together=set([('follower_content_type', 'follower_object_id', 'followee_content_type', 'followee_object_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('social_timeline', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='action',
            options={'ordering': ['-timestamp']},
        ),
        migrations.AlterField(
            model_name='action',
            name='timestamp',
            field=models.DateTimeField(default=django.utils.timezone.now),
            preserve_default=True,
        ),
    ]



from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone


class Follow(models.Model):
    follower_content_type = models.ForeignKey(ContentType)
    follower_object_id = models.PositiveIntegerField()
    follower = GenericForeignKey('follower_content_type', 'follower_object_id')
    followee_content_type = models.ForeignKey(
        ContentType, related_name="followee_follow_set")
    followee_object_id = models.PositiveIntegerField()
    followee = GenericForeignKey('followee_content_type', 'followee_object_id')
    is_mutual = models.BooleanField(default=False)

    class Meta:
        unique_together = (
            (
                'follower_content_type',
                'follower_object_id',
                'followee_content_type',
                'followee_object_id',
            ),
        )


class Action(models.Model):
    actor_content_type = models.ForeignKey(ContentType)
    actor_object_id = models.PositiveIntegerField()
    actor = GenericForeignKey('actor_content_type', 'actor_object_id')
    verb = models.CharField(max_length=255)
    target_content_type = models.ForeignKey(
        ContentType, blank=True, null=True, related_name="target_action_set")
    target_object_id = models.PositiveIntegerField(blank=True, null=True)
    target = GenericForeignKey('target_content_type', 'target_object_id')
    timestamp = models.DateTimeField(default=timezone.now)

    class Meta:
        ordering = ['-timestamp']

    def _get_link(self, obj):
        if obj is None:
            return
        if hasattr(obj, 'get_absolute_url'):
            return obj.get_absolute_url()

    def actor_link(self):
        return self._get_link(self.actor)

    def target_link(self):
        return self._get_link(self.target)

    @staticmethod
    def create_action(actor, verb, target=None, timestamp=None):
        action = Action(actor=actor, verb=verb)
        if target is not None:
            action.target = target
        if timestamp is not None:
            action.timestamp = timestamp
        action.save()
        return action

    @staticmethod
    def get_timeline(verb=None, count=None, time=None):
        actions = Action.objects.all()
        if verb:
            actions = actions.filter(verb=verb)
        if time:
            actions = actions.filter(timestamp__lte=time)
        if count is not None:
            actions = actions[:count]
        return actions



from setuptools import setup, find_packages

setup(
    name = "django-social-timeline",
    version = "0.2",
    author = "David Burke",
    author_email = "david@thelabnyc.com",
    description = ("Social networking for django. Following system and timeline of events."),
    license = "Apache License",
    keywords = ["django", "social networking"],
    url = "https://github.com/thelabnyc/django-social-timeline",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        'Environment :: Web Environment',
        'Framework :: Django',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: Apache Software License",
    ],
    install_requires=[
    ]
)



from django.contrib import admin
from .models import Action, Follow


@admin.register(Follow)
class FollowAdmin(admin.ModelAdmin):
    list_display = ('follower', 'followee')
    related_lookup_fields = {
        'generic': [
            ['follower_content_type', 'follower_object_id'],
            ['followee_content_type', 'followee_object_id'],
        ],
    }


@admin.register(Action)
class ActionAdmin(admin.ModelAdmin):
    list_display = ('actor', 'verb', 'target', 'timestamp')
    list_filter = ('actor_content_type', 'verb', 'timestamp',
                   'target_content_type')
    related_lookup_fields = {
        'generic': [
            ['actor_content_type', 'actor_object_id'],
            ['target_content_type', 'target_object_id'],
        ],
    }



from django.contrib.contenttypes.models import ContentType
from rest_framework import serializers
from .models import Action


class ContentTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ContentType


class StrRelatedField(serializers.Field):
    def to_representation(self, value):
        return str(value)


class ActionSerializer(serializers.ModelSerializer):
    actor_content_type = ContentTypeSerializer(read_only=True)
    actor = StrRelatedField(read_only=True)
    target_content_type = ContentTypeSerializer(read_only=True)
    target = StrRelatedField(read_only=True)

    class Meta:
        model = Action
        fields = (
            'timestamp', 'target_link', 'target', 'target_object_id',
            'target_content_type', 'verb', 'actor', 'actor_content_type',
            'actor_object_id', 'actor_link',
        )
